from bobocep.receiver.bobo_receiver import BoboReceiver
from bobocep.receiver.generators.data.bobo_null_data import BoboNullData
from bobocep.setup.task.bobo_task import BoboTask


class BoboNullDataGenerator(BoboTask):
    """
    A generator that periodically sends copies of some arbitrary data
    (i.e. *null data*) to a BoboReceiver instance.

    :param receiver: The data receiver to which null events are sent.
    :type receiver: BoboReceiver

    :param null_data: The null data to send.
    :type null_data: BoboNullData

    :param active: Whether task should start in an active state,
                   defaults to True.
    :type active: bool, optional
    """

    def __init__(self,
                 receiver: BoboReceiver,
                 null_data: BoboNullData,
                 active: bool = True) -> None:
        super().__init__(active=active)

        self.receiver = receiver
        self.null_data = null_data

    def _loop(self) -> None:
        self.receiver.add_data(self.null_data.get_null_data())

    def _setup(self) -> None:
        """"""

    def _cancel(self) -> None:
        """"""



from abc import ABC
from time import time_ns


class EpochNSClock(ABC):
    """A clock that returns the number of nanoseconds since the epoch."""

    @staticmethod
    def generate_timestamp() -> int:
        """
        :return: Nanoseconds since the epoch.
        """
        return time_ns()



from abc import ABC, abstractmethod

from bobocep.rules.events.action_event import ActionEvent
from bobocep.rules.events.composite_event import CompositeEvent


class IProducerSubscriber(ABC):
    """An interface to subscribe to Producer events."""

    @abstractmethod
    def on_accepted_producer_event(self, event: CompositeEvent):
        """
        Complex events accepted by a Producer.

        :param event: An accepted event.
        :type event: CompositeEvent
        """

    @abstractmethod
    def on_rejected_producer_event(self, event: CompositeEvent):
        """
        Complex events rejected by a Producer.

        :param event: A rejected event.
        :type event: CompositeEvent
        """

    @abstractmethod
    def on_producer_action(self, event: ActionEvent):
        """
        When an action has been attempted by a BoboAction instance subscribed
        to a Producer.

        :param event: The action event.
        :type event: ActionEvent
        """



from typing import List

from bobocep.rules.nfas.patterns.bobo_pattern_layer import BoboPatternLayer
from bobocep.rules.predicates.bobo_predicate import BoboPredicate


class BoboPattern:
    """A pattern that describes the structure of a nondeterministic finite
    automata, where each layer consists of a group of states under a common
    name."""

    def __init__(self) -> None:
        super().__init__()

        self.layers = []
        self.preconditions = []
        self.haltconditions = []

    def next(self,
             label: str,
             predicate: BoboPredicate,
             times: int = 1,
             loop: bool = False,
             optional: bool = False) -> 'BoboPattern':
        """Adds a new state, with strict contiguity.

        :param label: The label with which the state will be associated.
        :type label: str

        :param predicate: The predicate that the state will use for evaluation.
        :type predicate: BoboPredicate

        :param times: How many copies of the state to have in sequence,
                      default to 1.
        :type times: int, optional

        :param loop: Whether the state is self-looping, defaults to False.
        :type loop: bool, optional

        :param optional: Whether the state is optional, defaults to False.
        :type optional: bool, optional

        :return: The current pattern.
        """

        self.layers.append(BoboPatternLayer(label=label,
                                            predicates=[predicate],
                                            times=times,
                                            loop=loop,
                                            strict=True,
                                            negated=False,
                                            optional=optional))
        return self

    def not_next(self,
                 label: str,
                 predicate: BoboPredicate) -> 'BoboPattern':
        """Adds a new negated state, with strict contiguity.

        :param label: The label with which the state will be associated.
        :type label: str

        :param predicate: The predicate that the state will use for evaluation.
        :type predicate: BoboPredicate

        :return: The current pattern.
        """

        self.layers.append(BoboPatternLayer(label=label,
                                            predicates=[predicate],
                                            times=1,
                                            loop=False,
                                            strict=True,
                                            negated=True,
                                            optional=False))
        return self

    def followed_by(self,
                    label: str,
                    predicate: BoboPredicate,
                    times: int = 1,
                    loop: bool = False,
                    optional: bool = False) -> 'BoboPattern':
        """Adds a new state, with relaxed contiguity.

        :param label: The label with which the state will be associated.
        :type label: str

        :param predicate: The predicate that the state will use for evaluation.
        :type predicate: BoboPredicate

        :param times: How many copies of the state to have in sequence,
                      default to 1.
        :type times: int, optional

        :param loop: Whether the state is self-looping, defaults to False.
        :type loop: bool, optional

        :param optional: Whether the state is optional, defaults to False.
        :type optional: bool, optional

        :return: The current pattern.
        """

        self.layers.append(BoboPatternLayer(label=label,
                                            predicates=[predicate],
                                            times=times,
                                            loop=loop,
                                            strict=False,
                                            negated=False,
                                            optional=optional))
        return self

    def not_followed_by(self,
                        label: str,
                        predicate: BoboPredicate) -> 'BoboPattern':
        """Adds a new negated state, with relaxed contiguity.

        :param label: The label with which the state will be associated.
        :type label: str

        :param predicate: The predicate that the state will use for evaluation.
        :type predicate: BoboPredicate

        :return: The current pattern.
        """

        self.layers.append(BoboPatternLayer(label=label,
                                            predicates=[predicate],
                                            times=1,
                                            loop=False,
                                            strict=False,
                                            negated=True,
                                            optional=False))
        return self

    def followed_by_any(self,
                        label: str,
                        predicates: List[BoboPredicate]) -> 'BoboPattern':
        """Adds a new state for each predicate, with non-deterministic relaxed
        contiguity.

        :param label: The label with which the states will be associated.
        :type label: str

        :param predicates: The predicates that the states will use for
                           evaluation.
        :type predicates: List[BoboPredicate]

        :return: The current pattern.
        """

        self.layers.append(BoboPatternLayer(label=label,
                                            predicates=predicates,
                                            times=1,
                                            loop=False,
                                            strict=False,
                                            negated=False,
                                            optional=False))
        return self

    def precondition(self, predicate: BoboPredicate) -> 'BoboPattern':
        """Adds a new precondition predicate.

        :param predicate: The precondition predicate.
        :type predicate: BoboPredicate

        :return: The current pattern.
        """

        self.preconditions.append(predicate)
        return self

    def haltcondition(self, predicate: BoboPredicate) -> 'BoboPattern':
        """Adds a new haltcondition predicate.

        :param predicate: The haltcondition predicate.
        :type predicate: BoboPredicate

        :return: The current pattern.
        """

        self.haltconditions.append(predicate)
        return self

    def append(self, patterns: List['BoboPattern']) -> 'BoboPattern':
        """Appends a list of patterns to the current pattern, in list order.

        :param patterns: The patterns to append to the current pattern.
        :type patterns: List[BoboPattern]

        :return: The current pattern.
        """

        for pattern in patterns:
            self.layers.extend(pattern.layers)
            self.preconditions.extend(pattern.preconditions)
            self.haltconditions.extend(pattern.haltconditions)

        return self



from abc import abstractmethod
from threading import RLock

from bobocep.producer.producer_subscriber import IProducerSubscriber
from bobocep.receiver.clocks.epoch_ns_clock import EpochNSClock
from bobocep.rules.actions.action_subscriber import IActionSubscriber
from bobocep.rules.events.action_event import ActionEvent
from bobocep.rules.events.composite_event import CompositeEvent


class BoboAction(IProducerSubscriber):
    """A :code:`bobocep` action that can execute some given task.

    :param name: The action name, defaults to an empty string.
    :type name: str, optional
    """

    def __init__(self, name: str = None) -> None:
        super().__init__()

        self.name = name if name is not None else ""
        self._subs = []
        self._lock = RLock()

    @abstractmethod
    def _perform_action(self, event: CompositeEvent) -> bool:
        """"""

    def execute(self, event: CompositeEvent) -> ActionEvent:
        """
        Executes the action.

        :param event: An event to use as part of the action process.
        :type event: CompositeEvent

        :return: A Tuple containing the success of the action execution, and
                 any exception that was raised on failure, or None if an
                 exception is not raised.
        """

        with self._lock:
            success = False
            exception = None
            description = None

            try:
                success = self._perform_action(event=event)
            except Exception as e:
                exception = type(e).__name__
                description = str(e)

            action_event = ActionEvent(
                timestamp=EpochNSClock.generate_timestamp(),
                name=self.name,
                success=success,
                for_event=event,
                exception=exception if exception is not None else None,
                description=description if description is not None else None
            )

            self._notify_action_attempt(event=action_event)
            return action_event

    def on_accepted_producer_event(self, event: CompositeEvent) -> None:
        self.execute(event=event)

    def on_rejected_producer_event(self, event: CompositeEvent) -> None:
        """"""

    def on_producer_action(self, event: ActionEvent):
        """"""

    def subscribe(self, subscriber: IActionSubscriber) -> None:
        """
        :param subscriber: Subscribes to events from the Action.
        :type subscriber: IActionSubscriber
        """

        with self._lock:
            if subscriber not in self._subs:
                self._subs.append(subscriber)

    def unsubscribe(self, unsubscriber: IActionSubscriber) -> None:
        """
        :param unsubscriber: Unsubscribes to events from the Action.
        :type unsubscriber: IActionSubscriber
        """

        with self._lock:
            if unsubscriber in self._subs:
                self._subs.remove(unsubscriber)

    def _notify_action_attempt(self, event: ActionEvent):
        for sub in self._subs:
            sub.on_action_attempt(event=event)



from queue import Queue
from typing import List

from bobocep.decider.decider_subscriber import IDeciderSubscriber
from bobocep.decider.handlers.bobo_nfa_handler import BoboNFAHandler
from bobocep.decider.handlers.nfa_handler_subscriber import \
    INFAHandlerSubscriber
from bobocep.producer.producer_subscriber import IProducerSubscriber
from bobocep.receiver.receiver_subscriber import IReceiverSubscriber
from bobocep.rules.events.action_event import ActionEvent
from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory
from bobocep.setup.distributed.incoming.dist_incoming_subscriber import \
    IDistIncomingSubscriber
from bobocep.setup.task.bobo_task import BoboTask


class BoboDecider(BoboTask,
                  IReceiverSubscriber,
                  IProducerSubscriber,
                  INFAHandlerSubscriber,
                  IDistIncomingSubscriber):
    """A :code:`bobocep` data decider.

    :param max_queue_size: The maximum data queue size,
                           defaults to 0 (infinite).
    :type max_queue_size: int, optional

    :param active: Whether task should start in an active state,
                   defaults to True.
    :type active: bool, optional
    """

    def __init__(self, max_queue_size: int = 0, active: bool = True) -> None:
        super().__init__(active=active)

        self._event_queue = Queue(maxsize=max_queue_size)
        self._nfa_handlers = {}
        self._subs = {}

    def _loop(self) -> None:
        if not self._event_queue.empty():
            event = self._event_queue.get_nowait()

            if event is not None:
                for nfa_handler in self._nfa_handlers.values():
                    nfa_handler.process(event)

    def add_nfa_handler(self, nfa_handler: BoboNFAHandler) -> None:
        """
        Adds a new NFA handler to the decider.

        :param nfa_handler: The NFA handler to add.
        :type nfa_handler: BoboNFAHandler

        :raises RuntimeError: NFA handler name already in use.
        """

        with self._lock:
            if not self._is_cancelled:
                if nfa_handler.nfa.name not in self._nfa_handlers:
                    self._nfa_handlers[nfa_handler.nfa.name] = nfa_handler
                    nfa_handler.subscribe(self)
                else:
                    raise RuntimeError("NFA handler name {} already in use."
                                       .format(nfa_handler.nfa.name))

    def get_all_handlers(self) -> List[BoboNFAHandler]:
        """
        :return: A list of all handlers.
        """

        with self._lock:
            return list(self._nfa_handlers.values())

    def get_handler(self, name: str) -> BoboNFAHandler:
        """
        :param name: The handler name.
        :type name: str

        :return: The handler, or None if not found.
        """

        with self._lock:
            return self._nfa_handlers.get(name)

    def on_receiver_event(self, event: BoboEvent) -> None:
        if not self._is_cancelled:
            self._event_queue.put(event)

    def on_accepted_producer_event(self, event: CompositeEvent) -> None:
        if not self._is_cancelled:
            if event.name in self._nfa_handlers:
                self._nfa_handlers[event.name].add_recent(event)

            self._event_queue.put(event)

    def on_producer_action(self, event: ActionEvent):
        # producer actions are added to the handler's recent events
        if not self._is_cancelled:
            if event.for_event.name in self._nfa_handlers:
                self._nfa_handlers[event.for_event.name].add_recent(event)

    def on_handler_final(self,
                         nfa_name: str,
                         run_id: str,
                         event: CompositeEvent):
        # notify producer on a new complex event being identified
        with self._lock:
            if nfa_name in self._subs:
                for sub in self._subs[nfa_name]:
                    sub.on_decider_complex_event(event=event)

    def subscribe(self,
                  nfa_name: str,
                  subscriber: IDeciderSubscriber) -> None:
        """
        :param nfa_name: The NFA to subscribe to.
        :type nfa_name: str

        :param subscriber: Subscribes to NFA in Decider.
        :type subscriber: IDeciderSubscriber

        :raises RuntimeError: NFA name not found.
        """

        with self._lock:
            if not self._is_cancelled:
                if nfa_name not in self._nfa_handlers:
                    raise RuntimeError("NFA name {} not found in handlers."
                                       .format(nfa_name))

                if nfa_name not in self._subs:
                    self._subs[nfa_name] = []

                if subscriber not in self._subs[nfa_name]:
                    self._subs[nfa_name].append(subscriber)

    def unsubscribe(self,
                    nfa_name: str,
                    unsubscriber: IDeciderSubscriber) -> None:
        """
        :param nfa_name: The NFA to unsubscribe from.
        :type nfa_name: str

        :param unsubscriber: Subscribes to NFA in Decider.
        :type unsubscriber: IDeciderSubscriber
        """

        with self._lock:
            if not self._is_cancelled:
                if unsubscriber in self._subs[nfa_name]:
                    self._subs[nfa_name].remove(unsubscriber)

    def on_dist_run_transition(self,
                               nfa_name: str,
                               run_id: str,
                               state_name_from: str,
                               state_name_to: str,
                               event: BoboEvent):
        with self._lock:
            handler = self._nfa_handlers.get(nfa_name)

            if handler is None:
                raise RuntimeError(
                    "Handler not found for NFA {}.".format(nfa_name))

            handler.force_run_transition(
                run_id=run_id,
                state_name_from=state_name_from,
                state_name_to=state_name_to,
                event=event)

    def on_dist_run_clone(self,
                          nfa_name: str,
                          run_id: str,
                          next_state_name: str,
                          next_event: BoboEvent):
        with self._lock:
            handler = self._nfa_handlers.get(nfa_name)

            if handler is None:
                raise RuntimeError(
                    "Handler not found for NFA {}.".format(nfa_name))

            handler.force_run_clone(
                state_name=next_state_name,
                event=next_event,
                parent_run_id=run_id
            )

    def on_dist_run_halt(self,
                         nfa_name: str,
                         run_id: str):
        with self._lock:
            handler = self._nfa_handlers.get(nfa_name)

            if handler is None:
                raise RuntimeError(
                    "Handler not found for NFA {}.".format(nfa_name))

            handler.force_run_halt(
                run_id=run_id)

    def on_dist_run_final(self,
                          nfa_name: str,
                          run_id: str,
                          history: BoboHistory) -> None:
        with self._lock:
            handler = self._nfa_handlers.get(nfa_name)

            if handler is None:
                raise RuntimeError(
                    "Handler not found for NFA {}.".format(nfa_name))

            handler.force_run_final(
                run_id=run_id,
                history=history)

    def on_dist_action(self, event: ActionEvent) -> None:
        with self._lock:
            handler = self._nfa_handlers.get(event.for_event.name)

            if handler is None:
                raise RuntimeError("Handler not found for NFA {}.".format(
                    event.for_event.name))

            handler.add_recent(event)

    def _setup(self) -> None:
        """"""

    def _cancel(self) -> None:
        """"""

    def on_invalid_data(self, data):
        """"""

    def on_rejected_producer_event(self, event: CompositeEvent):
        """"""

    def on_handler_transition(self,
                              nfa_name: str,
                              run_id: str,
                              state_name_from: str,
                              state_name_to: str,
                              event: BoboEvent):
        """"""

    def on_handler_clone(self,
                         nfa_name: str,
                         run_id: str,
                         state_name: str,
                         event: BoboEvent):
        """"""

    def on_handler_halt(self,
                        nfa_name: str,
                        run_id: str):
        """"""



from typing import Dict

from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent


class ActionEvent(BoboEvent):
    """An action event.

    :param timestamp: The event timestamp indicating when it was first
                      generated.
    :type timestamp: int

    :param name: The action name.
    :type name: str

    :param success: :code:`True` if action was successfully executed,
                    :code:`False` otherwise.
    :type success: bool

    :param for_event: The event for which the action was triggered.
    :type for_event: CompositeEvent

    :param exception: The name of an exception, if one was raised during
                      action execution. Defaults to an empty string.
    :type exception: str, optional

    :param description: The description provided with an exception, if one was
                        raised. Defaults to an empty string.
    :type description: str, optional

    :param data: The event data, defaults to an empty dict.
    :type data: Dict[str, str], optional

    :param event_id: The event ID, defaults to a randomly generated ID.
    :type event_id: str, optional
    """

    NAME = "name"
    SUCCESS = "success"
    FOR_EVENT = "for_event"
    EXCEPTION = "exception"
    DESCRIPTION = "description"

    def __init__(self,
                 timestamp: int,
                 name: str,
                 success: bool,
                 for_event: CompositeEvent,
                 exception: str = None,
                 description: str = None,
                 data: Dict[str, str] = None,
                 event_id: str = None) -> None:
        super().__init__(timestamp=timestamp,
                         data=data,
                         event_id=event_id)

        self.name = name
        self.success = success
        self.for_event = for_event
        self.exception = exception if exception is not None else ""
        self.description = description if description is not None else ""

    def to_dict(self) -> dict:
        """
        :return: A dict representation of the object.
        """

        return {
            self.TIMESTAMP: self.timestamp,
            self.NAME: self.name,
            self.SUCCESS: self.success,
            self.FOR_EVENT: self.for_event.to_dict(),
            self.EXCEPTION: self.exception,
            self.DESCRIPTION: self.description,
            self.DATA: self.data,
            self.EVENT_ID: self.event_id
        }



from abc import ABC, abstractmethod


class IDistOutgoingSubscriber(ABC):

    @abstractmethod
    def on_sync(self) -> None:
        """When the :code:`bobocep` instance has synchronised with other
        instances."""



from abc import ABC, abstractmethod

from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory


class IRunSubscriber(ABC):

    @abstractmethod
    def on_run_transition(self,
                          run_id: str,
                          state_name_from: str,
                          state_name_to: str,
                          event: BoboEvent,
                          notify: bool):
        """Triggers a response when a state transition occurs in a run.

        :param run_id: The run ID.
        :type run_id: str

        :param state_name_from: The original state before the transition.
        :type state_name_from: str

        :param state_name_to: The new state after the transition.
        :type state_name_to: str

        :param event: The event that caused the state transition.
        :type event: BoboEvent

        :param notify: Whether to notify handler subscribers of the transition.
        :type notify: bool
        """

    @abstractmethod
    def on_run_clone(self,
                     state_name: str,
                     event: BoboEvent,
                     parent_run_id: str,
                     force_parent: bool,
                     notify: bool):
        """Triggers a response when a clone occurs in a run. Newly created runs
        are considered to be cloned but without a parent run.

        :param state_name: The state of the cloned run.
        :type state_name: str

        :param event: The event that caused the clone.
        :type event: BoboEvent

        :param parent_run_id: The parent run ID, defaults to None.
        :type parent_run_id: str, optional

        :param force_parent: Ensures that a run is found with the parent run
                             ID. Otherwise, an exception is raised.
        :type force_parent: bool

        :param notify: Whether to notify handler subscribers of the clone.
        :type notify: bool
        """

    @abstractmethod
    def on_run_final(self,
                     run_id: str,
                     history: BoboHistory,
                     notify: bool):
        """Triggers a response when a run reaches its final state.

        :param run_id: The run ID.
        :type run_id: str

        :param history: The history of events that led to run completion.
        :type history: BoboHistory

        :param notify: Whether to notify handler subscribers of the run
                       reaching its final state.
        :type notify: bool
        """

    @abstractmethod
    def on_run_halt(self,
                    run_id: str,
                    notify: bool):
        """Triggers a response when a run halts. This occurs if a run reaches
        its final state, or if a run halts prematurely.

        :param run_id: The run ID.
        :type run_id: str

        :param notify: Whether to notify handler subscribers of the halting.
        :type notify: bool
        """



from typing import Dict, List

from bobocep.rules.events.bobo_event import BoboEvent


class BoboHistory:
    """A :code:`bobocep` event history.

    :param events: The current history of events, where each key is a label
                   with which a list of events are associated,
                   defaults to an empty dict.
    :type events: Dict[str, List[BoboEvent]], optional
    """

    def __init__(self, events: Dict[str, List[BoboEvent]] = None) -> None:
        super().__init__()

        self.events = {} if events is None else events
        self.first = None
        self.last = None

        for event_list in self.events.values():
            for event in event_list:
                if self.first is None or \
                        event.timestamp < self.first.timestamp:
                    self.first = event

                if self.last is None or event.timestamp > self.last.timestamp:
                    self.last = event

    def to_dict(self) -> dict:
        """
        :return: A dict representation of the object.
        """

        dict_history = {}

        for key in self.events.keys():
            key_events = self.events[key]

            dict_history[key] = []

            for event in key_events:
                dict_history[key].append(event.to_dict())

        return dict_history



from typing import List

from bobocep.rules.actions.bobo_action import BoboAction
from bobocep.rules.actions.multi.multi_action import MultiAction
from bobocep.rules.events.composite_event import CompositeEvent


class SequentialAction(MultiAction):
    """An action that enables the sequential execution of multiple actions
    within a single BoboAction instance.

    :param actions: The actions to execute, in list order.
    :type actions: List[BoboAction]

    :param name: The action name, defaults to an empty string.
    :type name: str, optional

    :param all_success: All actions must be successful to be a successful
                        execution; otherwise, the success of *at least one*
                        action would result in an overall success,
                        Defaults to True.
    :type all_success: bool, optional

    :param early_stop: Will immediately return False on the first unsuccessful
                       action and not execute any remaining actions,
                       defaults to True. Note that any uncaught exception would
                       still cause an early stop and immediate failure.
    :type early_stop: bool, optional
    """

    def __init__(self,
                 actions: List[BoboAction],
                 name: str = None,
                 all_success: bool = True,
                 early_stop: bool = True) -> None:
        super().__init__(name=name)

        self._actions = actions
        self._all_success = all_success
        self._early_stop = early_stop

    def _perform_action(self, event: CompositeEvent) -> bool:
        results = []

        for action in self._actions:
            if action.execute(event).success:
                results.append(True)
            elif self._early_stop:
                return False
            else:
                results.append(False)

        return all(r is True for r in results) if self._all_success \
            else any(r is True for r in results)
